from .pos_enc import positional_encoding
from .segmented import segment_encoding, pool_batch, pool_sample
from .models import ChessBERT, ChessPooledBERT
from ._loss import loss
from .decode import argmax_ucis, greedy_pgn
from .data import Batch, Sample, parse_sample, collate_same, collate_separate

__all__ = [
  'positional_encoding', 'segment_encoding',
  'pool_batch', 'pool_sample',
  'ChessBERT', 'ChessPooledBERT', 'loss',
  'argmax_ucis', 'greedy_pgn',
  'Batch', 'Sample', 'parse_sample', 'collate_same', 'collate_separate'
]


from typing import Sequence
from jaxtyping import Float
import torch
import chess_utils as cu

def decode_file(file: int):
  return chr(file + ord('a'))

def decode_rank(rank: int):
  return str(rank + 1)

def decode_uci(from_file: int, from_rank: int, to_file: int, to_rank: int):
  return decode_file(from_file) + decode_rank(from_rank) + decode_file(to_file) + decode_rank(to_rank)

def argmax_ucis(logits: Float[torch.Tensor, 'B L 37']) -> Sequence[Sequence[str]]:
  batch_size = logits.size(0)
  from_files = torch.argmax(logits[:, :, 0:8].reshape(batch_size, -1, 8), dim=-1)
  from_ranks = torch.argmax(logits[:, :, 8:16].reshape(batch_size, -1, 8), dim=-1)
  to_files = torch.argmax(logits[:, :, 16:24].reshape(batch_size, -1, 8), dim=-1)
  to_ranks = torch.argmax(logits[:, :, 24:32].reshape(batch_size, -1, 8), dim=-1)
  
  return [
    [
      decode_uci(*[int(i.item()) for i in idxs])
      for idxs in zip(from_files[b], from_ranks[b], to_files[b], to_ranks[b])
    ]
    for b in range(batch_size)
  ]

def greedy_pgn(logits: Float[torch.Tensor, 'B L 37']) -> Sequence[Sequence[str]]:
  return [list(cu.ucis2sans(ucis)) for ucis in argmax_ucis(logits)]


from typing import Sequence, NamedTuple
from jaxtyping import Int
from torch import Tensor
import torch
from haskellian import iter as I
from .. import Char2Num, char2num as default_char2num, parse_sample as _parse_sample

class Sample(NamedTuple):
  input_ids: Int[Tensor, 'seq_len']
  word_ends: Sequence[int]
  labs: Int[Tensor, 'seq_len 5']

class Batch(NamedTuple):
  input_ids: Int[Tensor, 'batch seq_len']
  word_ends: Sequence[Sequence[int]]
  labs: Int[Tensor, 'batch seq_len 5']

def parse_sample(line: str, char2num: Char2Num = default_char2num) -> Sample:
  (tokens, ends), labs = _parse_sample(line)
  input_ids = [char2num[tok] for tok in tokens]
  return Sample(torch.tensor(input_ids), ends, torch.tensor(labs))

def collate_same(batch: Sequence[Sample], *, max_len: int, pad_token_id: int, ignore_idx: int = -100) -> Batch:
  """Pad the input_ids and labs to the same length. (Used by BERT1)"""

  input_ids, ends, labs = I.unzip(batch)
  batch_size = len(input_ids)
  max_len = min(max(len(x) for x in input_ids), max_len)
  padded_input_ids = torch.full((batch_size, max_len), fill_value=pad_token_id)
  padded_labs = torch.full((batch_size, max_len, 5), fill_value=ignore_idx)

  for i in range(batch_size):
    padded_input_ids[i, :len(input_ids[i])] = input_ids[i][:max_len]
    padded_labs[i, :len(labs[i])] = labs[i][:max_len]

  return Batch(padded_input_ids, ends, padded_labs)

def collate_separate(batch: Sequence[Sample], *, max_len: int, pad_token_id: int, ignore_idx: int = -100) -> Batch:
  """Pad input_ids and labels separately (used by pooled-BERT)"""

  input_ids, ends, labs = I.unzip(batch)
  batch_size = len(input_ids)
  max_input_len = min(max(len(x) for x in input_ids), max_len)
  max_lab_len = min(max(len(x) for x in labs), max_len)
  padded_input_ids = torch.full((batch_size, max_input_len), fill_value=pad_token_id)
  padded_labs = torch.full((batch_size, max_lab_len, 5), fill_value=ignore_idx)

  for i in range(batch_size):
    padded_input_ids[i, :len(input_ids[i])] = input_ids[i][:max_len]
    padded_labs[i, :len(labs[i])] = labs[i][:max_len]

  return Batch(padded_input_ids, ends, padded_labs)


from typing import Sequence, Iterable, NamedTuple, Literal, TypeAlias
import chess_utils as cu

class TokenizedInput(NamedTuple):
	tokens: Sequence[str | Literal['[SEP]']]
	ends: Sequence[int]

def tokenize(words: Sequence[str]) -> TokenizedInput:
	"""Split words into characters, adding '[SEP]' between words. `return.ends` are the original word end indices"""
	tokens = []
	ends = []
	end = 0
	for word in words:
		tokens.extend(word)
		tokens.append('[SEP]')
		end += len(word)
		ends.append(end)
	
	return TokenizedInput(tokens, ends)
	
promotion_ids = dict(n=1, b=2, r=3, q=4)
def encode_promotion(piece: Literal['n', 'b', 'r', 'q'] | None) -> int:
  return 0 if piece is None else promotion_ids[piece]

def encode_uci(e2e4q: str) -> tuple[int, int, int, int, int]:
  from_file = ord(e2e4q[0]) - ord('a')
  from_rank = int(e2e4q[1]) - 1
  to_file = ord(e2e4q[2]) - ord('a')
  to_rank = int(e2e4q[3]) - 1
  promotion = encode_promotion(e2e4q[4] if len(e2e4q) == 5 else None) # type: ignore
  return from_file, from_rank, to_file, to_rank, promotion

Label: TypeAlias = tuple[int, int, int, int, int]

def labels(ucis: Iterable[str]) -> Sequence[Label]:
	return [encode_uci(uci) for uci in ucis]

class Sample(NamedTuple):
	inputs: TokenizedInput
	labs: Sequence[Label]

def parse_sample(line: str) -> Sample:
	sans = line.strip('\n').replace('+', '').replace('#', '').split(' ')
	ucis = cu.sans2ucis(sans)
	inputs = tokenize(sans)
	labs = labels(ucis)
	return Sample(inputs, labs)


from typing import Sequence, Literal
from jaxtyping import Float
from haskellian import iter as I
import torch

def segment_encoding(
  seq_len: int, /, *, ends: Sequence[int],
  encoding: Float[torch.Tensor, 'max_len d_model']
) -> Float[torch.Tensor, 'seq_len d_model']:

  hidden_size = encoding.size(1)
  output = torch.zeros(seq_len, hidden_size, dtype=encoding.dtype, device=encoding.device)
  
  for i, (start, end) in I.pairwise([0, *ends]).enumerate():
    output[:, start:end] = encoding[:seq_len, i][..., None]

  return output

def pool_sample(
  embeds: Float[torch.Tensor, 'seq_len hidden_size'],
  word_ends: Sequence[int], *,
  num_moves: int,
  mode: Literal['mean', 'max'] = 'max'
) -> Float[torch.Tensor, 'num_moves hidden_size']:
  pool = torch.mean if mode == 'mean' else (lambda x, dim: torch.max(x, dim).values)
  output = torch.zeros(num_moves, embeds.size(1), dtype=embeds.dtype, device=embeds.device)
  for start, end in I.pairwise([0, *word_ends]):
    output[start:end] = pool(embeds[start:end], dim=0)
  return output

def pool_batch(
  embeds: Float[torch.Tensor, 'batch seq_len hidden_size'],
  word_ends: Sequence[Sequence[int]],
  *, mode: Literal['mean', 'max'] = 'max'
) -> Float[torch.Tensor, 'batch num_moves hidden_size']:
  
  max_moves = max(len(ends) for ends in word_ends)
  return torch.stack([
    pool_sample(embed, ends, mode=mode, num_moves=max_moves)
    for embed, ends in zip(embeds, word_ends)
  ])


from jaxtyping import Int, Float
from torch import Tensor, nn

def loss(
  logits: Float[Tensor, 'batch seq_len 37'],
  labels: Int[Tensor, 'batch seq_len'], *,
  ce_loss: nn.CrossEntropyLoss = nn.CrossEntropyLoss(ignore_index=-100)
) -> Float[Tensor, 'batch']:
  """Cross-entropy loss across the separate categories
  - `logits[..., 0:8]`: start_file
  - `logits[..., 8:16]`: start_rank
  - `logits[..., 16:24]`: end_file
  - `logits[..., 24:32]`: end_rank
  - `logits[..., 32:37]`: promotion
  """
  split_logits = [
    logits[:, :, 0:8] .reshape(-1, 8),
    logits[:, :, 8:16].reshape(-1, 8),
    logits[:, :, 16:24].reshape(-1, 8),
    logits[:, :, 24:32].reshape(-1, 8),
    logits[:, :, 32:37].reshape(-1, 5),
  ]

  split_labs = [
    labels[:, :, 0].reshape(-1),
    labels[:, :, 1].reshape(-1),
    labels[:, :, 2].reshape(-1),
    labels[:, :, 3].reshape(-1),
    labels[:, :, 4].reshape(-1),
  ]
  losses = [ce_loss(logits, labs) for logits, labs in zip(split_logits, split_labs)]
  return sum(losses) # type: ignore


from typing import Mapping, Literal, TypeAlias
import string

SpecialToken: TypeAlias = Literal['[PAD]', '[SEP]']
Char2Num: TypeAlias = Mapping[str | SpecialToken, int]
Num2Char: TypeAlias = Mapping[int, str | SpecialToken]

VOCABULARY = string.ascii_letters + string.digits + '-='
char2num: Char2Num = { char: i for i, char in enumerate(VOCABULARY) } | {
  '[PAD]': len(VOCABULARY),
  '[SEP]': len(VOCABULARY) + 1
}
num2char: Num2Char = { i: char for char, i in char2num.items() }


from typing import Sequence
from jaxtyping import Int, Float
import torch
from torch import nn, Tensor
from transformers import BertConfig, BertModel
from ..pos_enc import positional_encoding
from ..segmented import segment_encoding

class ChessBERT(nn.Module):

  pos_enc: Float[Tensor, 'd_model max_len']
  
  def __init__(
    self, *,
    vocab_size: int,
    max_len: int,
    hidden_size: int = 768,
    attention_heads: int = 12
  ):
    super().__init__()
    # BERT configuration
    self.config = BertConfig(
      vocab_size=vocab_size,
      hidden_size=hidden_size,
      num_attention_heads=attention_heads,
      max_position_embeddings=max_len
    )
    self.register_buffer('pos_enc', positional_encoding(max_len, hidden_size))
    self.embedding = nn.Embedding(vocab_size, hidden_size)
    self.bert = BertModel(self.config)
    self.fc = nn.Linear(hidden_size, 4 * 8 + 5)

  def forward(
    self, input_ids: Int[Tensor, 'batch seq_len'],
    word_ends: Sequence[Sequence[int]]
  ) -> Float[Tensor, 'batch seq_len 37']:
    pos_enc = torch.stack([
      segment_encoding(input_ids.size(1), ends=ends, encoding=self.pos_enc)
      for ends in word_ends
    ])
    embedded: Float[Tensor, 'batch seq_len hidden_size'] = self.embedding(input_ids)
    x: Float[Tensor, 'batch seq_len hidden_size'] = embedded + pos_enc
    y = self.bert(inputs_embeds=x)
    seq_output = y.last_hidden_state
    return self.fc(seq_output)



from typing import Sequence

def sans_accuracy(labs: Sequence[str], preds: Sequence[str]) -> float:
  """Proportion of correct predictions."""
  return sum(l == p for l, p in zip(labs, preds)) / len(labs)


from typing import Sequence, Literal
from jaxtyping import Int, Float
from torch import nn, Tensor
from transformers import BertConfig, BertModel
from ..pos_enc import positional_encoding
from ..segmented import pool_batch

class ChessPooledBERT(nn.Module):

  pos_enc: Float[Tensor, 'd_model max_len']
  pool: Literal['mean', 'max']
  
  def __init__(
    self, *,
    vocab_size: int,
    max_len: int,
    pool: Literal['mean', 'max'] = 'max',
    pre_nheads: int = 12,
    pre_layers: int = 2,
    pre_ff_dims: int = 2048,
    hidden_size: int = 768,
    attention_heads: int = 12,
  ):
    super().__init__()
    # BERT configuration
    self.config = BertConfig(
      vocab_size=vocab_size,
      hidden_size=hidden_size,
      num_attention_heads=attention_heads,
      max_position_embeddings=max_len
    )
    self.pool = pool
    self.register_buffer('pos_enc', positional_encoding(max_len, hidden_size))
    self.pre_layers = nn.TransformerEncoder(
      nn.TransformerEncoderLayer(hidden_size, nhead=pre_nheads, dim_feedforward=pre_ff_dims, batch_first=True),
      num_layers=pre_layers
    )
    self.embedding = nn.Embedding(vocab_size, hidden_size)
    self.bert = BertModel(self.config)
    self.fc = nn.Linear(hidden_size, 4 * 8 + 5)

  def forward(
    self, input_ids: Int[Tensor, 'batch seq_len'],
    word_ends: Sequence[Sequence[int]]
  ) -> Float[Tensor, 'batch seq_len 37']:
    
    embeds: Float[Tensor, 'batch seq_len hidden_size'] = self.embedding(input_ids)
    enhanced_embeds: Float[Tensor, 'batch seq_len hidden_size'] = self.pre_layers(embeds)
    pooled_embeds = pool_batch(enhanced_embeds, word_ends, mode=self.pool)
    move_embeds: Float[Tensor, 'batch moves hidden_size'] = self.pos_enc[None, :len(pooled_embeds[1]), :] + pooled_embeds

    y = self.bert(inputs_embeds=move_embeds)
    seq_output = y.last_hidden_state
    return self.fc(seq_output)



from .samples import tokenize, labels, TokenizedInput, parse_sample, Sample
from .metrics import sans_accuracy
from ._vocab import char2num, num2char, VOCABULARY, Char2Num, Num2Char, SpecialToken
from . import pytorch

__all__ = [
  'tokenize', 'labels', 'TokenizedInput', 'parse_sample', 'Sample',
  'sans_accuracy',
  'char2num', 'num2char', 'VOCABULARY', 'Char2Num', 'Num2Char', 'SpecialToken',
  'pytorch',
]


import math
from jaxtyping import Float
import torch

def positional_encoding(max_len: int, d_model: int) -> Float[torch.Tensor, 'max_len d_model']:
  position = torch.arange(max_len)[:, None]
  div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
  pe = torch.zeros(max_len, d_model)
  pe[:, 0::2] = torch.sin(position * div_term)
  pe[:, 1::2] = torch.cos(position * div_term)
  return pe
